/*********************************************************************************
*                                                                                *
*   ODBPPP.h -- ObjectDatabase++ definitions for the ODBPP Class API functions   *
*                                                                                *
*   Copyright 2003-2020, Ekky Software Pty Ltd. All rights reserved.             *
*   http://www.ekkysoftware.com                                                  *
*                                                                                *
*********************************************************************************/

#pragma once

#ifndef _WINDEF_

typedef struct tagRECT{
   int left;                      // Specifies the x-coordinate of the upper-left corner of the rectangle.
   int top;                       // Specifies the y-coordinate of the upper-left corner of the rectangle.
   int right;                     // Specifies the x-coordinate of the lower-right corner of the rectangle
   int bottom;                    // Specifies the y-coordinate of the lower-right corner of the rectangle.
} RECT, *LPRECT;

typedef struct tagPOINT{
    int  x;
    int  y;
} POINT, *LPPOINT;

typedef struct _FILETIME {
   unsigned int dwLowDateTime;
   unsigned int dwHighDateTime;
} FILETIME, *LPFILETIME;

#endif //_WINDEF_

typedef struct tagCUBE{
   int left;                      // Specifies the x-coordinate of the upper-left-front corner of the cube.
   int top;                       // Specifies the y-coordinate of the upper-left-front corner of the cube.
   int front;                     // Specifies the z-coordinate of the upper-left-front corner of the cube.
   int right;                     // Specifies the x-coordinate of the lower-right-rear corner of the cube
   int bottom;                    // Specifies the y-coordinate of the lower-right-rear corner of the cube.
   int rear;                      // Specifies the z-coordinate of the lower-right-rear corner of the cube.
} CUBE, *LPCUBE;

typedef struct tagPOINT3D{
    int  x;
    int  y;
    int  z;
} POINT3D, *LPPOINT3D;

#ifdef _WINDOWS_
#  ifdef ODBPP_EXPORTS
#     define ODBPP_API            __declspec(dllexport)
#  else
#     define ODBPP_API            __declspec(dllimport)
#  endif
#else //!_WINDOWS_
   // For Linux users gcc -std=c++0x
#  if defined(_SOLARIS) // For Solaris users CC -xustr=ascii_utf16_ushort
      typedef unsigned short      char16_t;
#  endif //defined(_SOLARIS)
#  if defined(__GNUC__) && __GNUC__ >= 4
#     define ODBPP_API            __attribute__((visibility("default")))
#  else
#     define ODBPP_API          
#  endif
   typedef unsigned char          BYTE;
   typedef unsigned char*         LPBYTE;
#  define __int64                 long long int
#  define NO_ERROR                0
#endif //!_WINDOWS_

class ODBPP_API CODBPP{
public:
   enum DataType{
      BOOL                        = 3, // boolean
      CHAR8                       = 5, // 8-bit character
      CHAR16                      = 7, // 16-bit wide character
      INT8                        = 9, // 8-bit signed integer
      UINT8                       = 11,// 8-bit unsigned integer
      INT16                       = 13,// 16-bit signed integer
      UINT16                      = 15,// 16-bit unsigned integer
      INT32                       = 17,// 32-bit signed integer
      UINT32                      = 19,// 32-bit unsigned integer
      INT64                       = 21,// 64-bit signed integer
      UINT64                      = 23,// 64-bit unsigned integer
      FLOAT32                     = 25,// 32-bit floating point
      FLOAT64                     = 27,// 64-bit floating point
      DATE                        = 31,// 32-bit day counter from 1/1/1601
      FILETIME                    = 33,// 64-bit 100-nanosecond (10x-7) counter from 1/1/1601
      OBJECTID                    = 35,// 32-bit unsigned integer object pointer
      OBJECTID64                  = 43,// 64-bit integer object pointer, added Version 3.0
      ARRAY                       = 37,// unsigned char array for general use - indexable
      RECT                        = 39,// rectangle - indexable (S-Tree)
      CUBE                        = 41,// cube - indexable (S-Tree)
      ASTR                        = 2, // ASCII character string - null terminating UTF-8
      USTR                        = 4, // wide character string - null terminating UTF-16
      SBSTR                       = 6, // short binary string, first 2 bytes have the length of the SBSTR including the 2 bytes - non-indexable
      BSTR                        = 8, // binary string, first 4 bytes have the length of the BSTR including the 4 bytes - non-indexable
      PATTERN2D                   = 12,// binary string containing POINT tuples, added Version 4.1
      PATTERN3D                   = 14,// binary string containing POINT3D tuples, added Version 4.1
      SUB_TABLE                   = 10 // binary string containing sub table tuples, added Version 2.1
   };
   enum IndexType{
      B_TREE                      = 0, // basic B+Tree index
      S_TREE                      = 1, // Spatial Tree is as defined as 2D & 3D R-Tree
      T_TREE                      = 4, // Temporal Tree, added Version 3.5
      A_LIST                      = 2, // ASCII token list search, added Version 3.1
      W_LIST                      = 3, // Wide charater token list search, added Version 3.1
      PATTERN                     = 5  // 2 or 3 dimensional pattern, added Version 4.1
   };
   enum Mode{
      NONE                        = 0, // No mode
      DESCENDING                  = 1, // Used for index mode - all types - reverses the value of the field
      INVERTED                    = 2, // Used for index mode - ASTR & USTR types only - Inverts the string so that "abc" becomes "cba"
      TO_UPPER_CASE               = 4, // Used for index mode - ASTR & USTR types only - changes the casing to upper case
      TO_LOWER_CASE               = 8, // Used for index mode - ASTR & USTR types only - changes the casing to lower case - used when there are more lower case character than upper case
      TOKEN_FIELD                 = 16,// Used for index mode - ASTR, USTR, SBSTR, BSTR & ARRAY types only - used in token list indexes for the field which are to be tokenized
      REMOVE_WHITE_SPACES         = 32,// Used for index mode - ASTR & USTR types only - removes white spaces from the string so that "a b\tcd\r\nef" becomes "abcdef"
      ALPHANUMERIC_ONLY           = 64,// Used for index mode - ASTR & USTR types only - indexes on alpha and numeric characters only, for example "a-b\tcd*&4f" becomes "a b cd 4f", from v4.3
      NATURALIZED_16              = 128,// Used for index mode - ASTR & USTR types only - transforms numbers into 16 bit binary, for example "a128" becomes "a\x00\x80", from v4.3
      NATURALIZED_32              = 256,// Used for index mode - ASTR & USTR types only - transforms numbers into 32 bit binary, for example "a128" becomes "a\x00\x00\x00\x80", from v4.3
      NATURALIZED_64              = 384,// Used for index mode - ASTR & USTR types only - transforms numbers into 64 bit binary, for example "a128" becomes "a\x00\x00\x00\x00\x00\x00\x00\x80", from v4.3
      SHARED                      = 33,// Used when opening a database - to allow other transaction share access
      EXCLUSIVE                   = 34,// Used when opening a database - to allow no other transaction - required for some methods
      LOCK_FREE                   = 35,// Used when opening a database - Allows all transactions to operate concurrently without locking and only allows read only operations
      //Should there be?? READ_ONLY = 36,// Used when opening a database - Allows all transactions to operate concurrently and only allows read only operations
      EQUALTO                     = 40,// Used for index searches - finds equal to the given index value
      LESS_THAN                   = 41,// Used for index searches - finds less than the given index value
      GREATER_THAN                = 42,// Used for index searches - finds greater than the given index value
      LESS_EQUALTO                = 43,// Used for index searches - finds less than or equal to the given index value
      GREATER_EQUALTO             = 44,// Used for index searches - finds greater than or equal to the given index value
      FIRST                       = 45,// Used for index searches - finds first Object within the index
      LAST                        = 46,// Used for index searches - finds last Object withint the index
      NEXT                        = 47,// Used for index searches - finds next Object to the currently selected Object
      PREVIOUS                    = 48,// Used for index searches - finds previous Obejct to the currently selectes Object
      APPEND_RESULTS              = 16,// Used for index searches - allows SearchIndex method to append the results to previous SearchIndex results
      ACQUIRE_LOCKS               = 32,// Used for index searches - allows SearchIndex method to acquire locks before the ReadObject method
      NO_LOCK                     = 64,// Used for index searches - allow ReadObject method to operate without locking
      NO_WAIT                     = 128,// Used for index searches - allow ReadObject and DeleteObject method to return without waiting for locking
      SHORT_WAIT                  = 256 // Used for index searches - allow ReadObject and DeleteObject method to return after a short waiting for locking
   };
   enum Parameter {
       CYCLICAL_LOGGING           = 1,// Boolean to change between cyclical and instant logging
       CYCLICAL_LOG_SIZE          = 2,// Short for minimum size in Megabytes before moving on to the next file, 0 = default 32 MB
       CYCLICAL_LOG_HOURS         = 3,// Short for minimum hours before deleteing log file, 0 = default 120 hours
       CYCLICAL_CURR_MIN          = 4,// Read only short for current minimun cyclical log file
       CYCLICAL_CURR_MAX          = 5,// Read only short for current maximun cyclical log file
       REMOVE_ERRORING_OBJECTS    = 0x400// Force SetTableSchema to complete with duplicate index values
   };
   enum {
      ODBPP_UNREGISTERED          = 1,
      FILE_NOT_LOCAL              = 2,
      DATABASE_NOT_OPEN           = 3,
      DATABASE_NOT_EXCLUSIVE      = 4,
      DATABASE_NOT_EDITABLE       = 5,
      TABLE_NOT_OPEN              = 6,
      INVALID_PARAMETER           = 7,
      INVALID_OPERATION           = 8,
      INVALID_TABLE_NAME          = 9,
      INVALID_TABLE_ID            = 10,
      INVALID_TABLE_LOCK_ORDER    = 11,
      INVALID_FIELD_TYPE          = 12,
      INVALID_FIELD_ORDER         = 13,
      INVALID_FIELD_POS_IN_INDEX  = 14,
      INVALID_FIELD_TYPE_IN_INDEX = 15,
      INVALID_FIELD_NAME          = 42,
      INVALID_FIELD_LENGTH        = 66,
      INVALID_SEGMENT_MODE        = 16,
      INVALID_SEGMENT_COUNT       = 17,
      INVALID_SEGMENT_MULTIENTRY  = 61,
      INVALID_SEARCH_MODE         = 18,
      INVALID_SEARCH_STRING       = 63,
      INVALID_INDEX_TYPE          = 19,
      INVALID_INDEX_MAX_COUNT     = 20,
      INVALID_INDEX_LOCK_ORDER    = 26,
      INVALID_VERSION             = 70,
      INVALID_CONTROL_VERSION     = 21,//should automatically upgrade to newer version
      INVALID_TABLE_VERSION       = 22,
      INVALID_INDEX_VERSION       = 23,
      INVALID_LOG_VERSION         = 24,
      INVALID_LOG_POSITION        = 68,
      UNABLE_TO_LOCK_DATABASE     = 25,
      UNABLE_TO_LOCK_OBJECT       = 27,
      UNABLE_TO_LOCK_TRANSACTION  = 67,
      NOT_FOUND_TABLE             = 30,
      NOT_FOUND_OBJECT            = 31,
      NOT_FOUND_INDEX_KEY         = 32,
      NOT_FOUND_LOG_ENTRY         = 33,
      ALREADY_IN_USE_DATABASE     = 34,
      ALREADY_IN_USE_TABLE        = 35,
      ALREADY_IN_USE_OBJECT       = 36,
      ALREADY_IN_USE_INDEX_KEY    = 37,
      ALREADY_IN_USE_LOG_KEY      = 38,
      NO_VALID_OBJECT             = 39,
      NO_VALID_SEARCH_TREE        = 40,
      TRANSACTION_INCOMPLETE      = 41,
      CAST_UNAVAILABLE            = 43,
      INCONSISTENT_INDEX          = 44,
      INCONSISTENT_THREAD         = 28,
      CORRUPT_OBJECT_LENGTH       = 45,
      CORRUPT_OBJECT_FORMAT       = 46,
      CORRUPT_DELETE_ID_LIST      = 47,
      CORRUPT_FREE_SPACE_LIST     = 48,
      CORRUPT_B_TREE_VAR_LEAF     = 50,
      CORRUPT_T_LIST_BUCKET       = 62,
      CORRUPT_B_TREE              = 51,
      CORRUPT_S_TREE              = 52,
      CORRUPT_T_TREE              = 64,
      CORRUPT_PATTERN             = 65,
      CORRUPT_T_LIST              = 53,
      CORRUPT_CONTROL_FILE        = 54,
      CORRUPT_TABLE_FILE          = 55,
      CORRUPT_LOG_FILE            = 56,
      CORRUPT_LOG_ENTRY           = 69,
      CORRUPT_INDEX_FILE          = 57,
      CORRUPT_MEMORY_FILE         = 58,
      INTERNAL_ERROR              = 59,
      AES_ERROR                   = 60
   };
   class Segment{
   public:
      int field;                 // -1 for objectID, -2 for multi-entry row count, else array index Schema.Fields
      enum Mode mode;            // for B+-Tree     - ((ascending XOR descending) OR (to upper case XOR to lower case) OR remove white spaces)
                                 // for Token Lists - (token field OR (to upper case XOR to lower case))
   };
   class Index{
   public:
      unsigned char flags;       // 0x01 = variable, 0x02 = multi-entry, 0x04 = non-memory
      unsigned char type;        // enum IndexType
      unsigned short segmentCount;
      union{                     // For Solaris users CC -features=extensions for anonymous unions
         unsigned int length;    // for CODBPP::B_TREE, CODBPP::S_TREE and CODBPP::T_TREE - length << 2 = bytes
         struct{                 // for CODBPP::A_LIST and CODBPP::W_LIST -
            unsigned short binaryLength;// the length in bytes of the pre-token binary section
            unsigned short maxObjects;// maximum object count before disregarding token value, pow(2,maxObjects)
         };
      };
      Segment *segments;
   };
   class Field{
   public:
      enum DataType type;
      unsigned int offset;       // for variable field is the total length of all the fixed fields
      unsigned int length;       // length = bytes
      //unsigned int subFieldLength;// for CODBPP::SUB_TABLE, counts the fields in just this sub table, including any other nested subfields
      unsigned int subFieldCount;// for CODBPP::SUB_TABLE, counts all the fields in this sub table and any sub table within
      const char16_t *name;      // changed to utf-16 character version 4.0
      Field *subFields;
   };
   class Schema{
   public:
      unsigned short indexLockOrder;
      unsigned short indexCount;
      unsigned int tableID;
      unsigned int tableLockOrder;
      unsigned int fieldCount;
      //unsigned int maxField;// potentially add a total field count including all subfields
      unsigned int flags;        // 0x01 - 64 bit Object IDs, 0 = no, 1 = yes
                                 // 0x02 - Inverted index lock order, 0 = no, 1 = yes
      const char16_t *tableName; // changed to utf-16 character version 4.0
      const char16_t *indexPath; // changed to utf-16 character version 4.0, this sould be for the database and not individual tables
      Field *fields;
      Index *indexes;
   };
   class DatabaseMember{
   public:
      unsigned int tableID;
      unsigned int tableLockOrder;
      char16_t *tableName;       // changed to utf-16 character version 4.0
   };
   class Object{
   public:
      unsigned short MSN;
      unsigned int length;
      union{
         unsigned int ID;
         __int64 ID64;
      };
      unsigned char *fixed;
      unsigned char *variable;
   };
private:
   void *database, *tables;
public:
   CODBPP();
   ~CODBPP();
   unsigned __int64 AbortTransaction(bool releaseObjectLocks = true);
   unsigned __int64 AddObject(unsigned int tableID, CODBPP::Object* object = NULL, __int64 objectID = 0);
   unsigned __int64 AddTable(const char16_t* tableName, unsigned int tableID = 0, const char16_t* indexPath = NULL, unsigned int tableLockOrder = 0);
   unsigned __int64 AppendTransaction(unsigned int logNumber, __int64 position, unsigned int transactionID, __int64 length, const void* buffer);
   unsigned __int64 BeginTransaction(enum CODBPP::Mode mode = CODBPP::SHARED, unsigned int milliseconds = 60000);
   unsigned __int64 CloseDatabase();
   unsigned __int64 CloseTable(unsigned int tableID);
   unsigned __int64 CommitTransaction(bool releaseObjectLocks = true);
   unsigned __int64 CreateDatabase(const char16_t* controlFileName);
   unsigned __int64 CreateTable(const CODBPP::Schema *schema, unsigned int *tableID = NULL);
   unsigned __int64 DeleteObject(unsigned int tableID, __int64 objectID = 0, CODBPP::Object *object = NULL){ return DeleteObject(tableID,CODBPP::EQUALTO,object,0,&objectID);}
   unsigned __int64 DeleteObject(unsigned int tableID, enum CODBPP::Mode operation, CODBPP::Object *object = NULL, unsigned int index = 0, const void *key = NULL, unsigned int *matchingRow = NULL);
   unsigned __int64 EndTransaction();
   unsigned __int64 GetDatabaseMembers(CODBPP::DatabaseMember *members, unsigned int *count);
   unsigned __int64 GetDatabaseParameter(enum CODBPP::Parameter parameter, __int64* value);
   unsigned __int64 GetErrorMessage(const char16_t **message);
   unsigned __int64 GetTableSchema(unsigned int tableID, CODBPP::Schema *schema, unsigned int *length = NULL);
   unsigned __int64 GetVariableField(unsigned int tableID, unsigned int variableField, void **returnPosition, CODBPP::Object *object = NULL);
   unsigned __int64 GetVariableField(unsigned int tableID, unsigned int variableField, unsigned int subField, unsigned int row, void **returnPosition, CODBPP::Object *object = NULL);
   unsigned __int64 NewObject(unsigned int tableID, CODBPP::Object *object = NULL, unsigned int variableLength = 0);
   unsigned __int64 OpenDatabase(const char16_t *controlFileName);
   unsigned __int64 OpenTable(unsigned int tableID);
   unsigned __int64 PutVariableValues(unsigned int tableID, const void* variableBuffer, unsigned int variableLength /*in bytes, 0 for ignore*/, unsigned int variableStartField, unsigned int variableFieldCount, CODBPP::Object *object = NULL);
   unsigned __int64 PutVariableValues(unsigned int tableID, CODBPP::Object *object = NULL, unsigned int variableLength = 0);
   unsigned __int64 ReadObject(unsigned int tableID, __int64 objectID, CODBPP::Object *object = NULL){ return ReadObject(tableID,CODBPP::EQUALTO,object,0,&objectID);}
   unsigned __int64 ReadObject(unsigned int tableID, __int64 objectID, enum CODBPP::Mode operation, CODBPP::Object *object = NULL){ return ReadObject(tableID,(CODBPP::Mode)(CODBPP::EQUALTO|operation),object,0,&objectID);}
   unsigned __int64 ReadObject(unsigned int tableID, enum CODBPP::Mode operation, CODBPP::Object *object = NULL, unsigned int index = 0, const void *key = NULL, unsigned int *matchingRow = NULL);
   unsigned __int64 RemoveTable(unsigned int tableID, bool deleteFiles = true);
   unsigned __int64 RestoreObject(unsigned int tableID, __int64 objectID = 0, bool releaseObjectLock = true);
   unsigned __int64 RewriteObject(unsigned int tableID, CODBPP::Object *object = NULL);
   unsigned __int64 SearchIndex(unsigned int tableID, unsigned int index, const void *binaryKey, const void *tokenKey = NULL, enum CODBPP::Mode operation = CODBPP::NONE, __int64 *objectCount = NULL);
   unsigned __int64 SearchIndex(unsigned int tableID, unsigned int index, const void *key, unsigned int errorMargin, enum CODBPP::Mode operation = CODBPP::NONE, __int64 *objectCount = NULL);
   unsigned __int64 SetDatabaseParameter(enum CODBPP::Parameter parameter, __int64 value);
   unsigned __int64 SetTableSchema(unsigned int tableID, const CODBPP::Schema *schema = NULL, const unsigned int *fieldMapping = NULL, enum CODBPP::Parameter parameter = (CODBPP::Parameter)0);
   unsigned __int64 WaitNextCommit(unsigned int* logNumber, __int64* position, unsigned int* transactionID, __int64* length, void* buffer = NULL, unsigned int milliseconds = 15000 /*or INFINITE*/);
   static unsigned __int64 GetLicenseInformation(LPFILETIME start, LPFILETIME finish, unsigned int *installations = NULL, const char16_t **registeredName = NULL);
   static unsigned __int64 RegisterODBPP(const char16_t *registeredName, unsigned __int64 code1, unsigned __int64 code2);
   static unsigned __int64 Version();
};

